/**
 * This work was created by participants in the DataONE project, and is
 * jointly copyrighted by participating institutions in DataONE. For
 * more information on DataONE, see our web site at http://dataone.org.
 *
 *   Copyright ${year}
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.dataone.audit;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;
import org.dataone.client.D1Node;
import org.dataone.client.audit.AuditCNode;
import org.dataone.service.exceptions.InsufficientResources;
import org.dataone.service.exceptions.InvalidRequest;
import org.dataone.service.exceptions.InvalidToken;
import org.dataone.service.exceptions.NotAuthorized;
import org.dataone.service.exceptions.NotImplemented;
import org.dataone.service.exceptions.ServiceFailure;
import org.dataone.service.types.v1.AccessRule;
import org.dataone.service.types.v1.Event;
import org.dataone.service.types.v1.Log;
import org.dataone.service.types.v1.ObjectFormatIdentifier;
import org.dataone.service.types.v1.ObjectList;
import org.dataone.service.types.v1.Subject;
import org.dataone.service.util.D1Url;


/**
 * Utilities that are useful for generating test data.
 */
public class APITestUtils {

	static Logger logger = Logger.getLogger(APITestUtils.class.getName());

	public static ObjectList pagedListObjects(D1Node d1Node, Date fromDate, Date toDate, 
		      ObjectFormatIdentifier formatid, Boolean replicaStatus, Integer start, Integer count) 
		    throws InvalidRequest, InvalidToken, NotAuthorized, NotImplemented, ServiceFailure
		    {
		return APITestUtils.pagedListObjects(d1Node, fromDate, toDate, formatid, replicaStatus, start, count, null);
	
		    }
	
	/**
	 * Calls list objects iteratively using the paging mechanism in order to
	 * respect any server-imposed paging limits
	 * @param d1Node
	 * @param fromDate
	 * @param toDate
	 * @param formatid
	 * @param replicaStatus
	 * @param start
	 * @param count
	 * @param pageSize
	 * @return
	 * @throws InvalidRequest
	 * @throws InvalidToken
	 * @throws NotAuthorized
	 * @throws NotImplemented
	 * @throws ServiceFailure
	 */
	public static ObjectList pagedListObjects(D1Node d1Node, Date fromDate, Date toDate, 
      ObjectFormatIdentifier formatid, Boolean replicaStatus, Integer start, Integer count, Integer pageSize) 
    throws InvalidRequest, InvalidToken, NotAuthorized, NotImplemented, ServiceFailure
    {

		// if count is zero or less, set it to zero, and don't need to page
		if (count != null && count <= 0) {
			return d1Node.listObjects(null, fromDate, toDate, formatid, replicaStatus, start, 0);
		} 
		
		// use the pageSize to set the desired pageSize, it might not be respected
		ObjectList ol = d1Node.listObjects(null, fromDate, toDate, formatid, replicaStatus, start, pageSize);
		if (ol.getTotal() == ol.sizeObjectInfoList()) {
			// don't need to ask for more
			// but we might have retrieved more than we asked for in count parameter
			if (count != null && ol.sizeObjectInfoList() > count) {
				// need to trim the object list to match the requested amount
				ol.setObjectInfoList(ol.getObjectInfoList().subList(0, count));
				ol.setCount(count);
				return ol;
			}
		}
		
		// now determine if we have enough
		count = -1;
		int retrieved = ol.sizeObjectInfoList();
		int serverPageSize = ol.sizeObjectInfoList();  // server is happy to return this amount at a time.
		int totalNeeded = count > ol.getTotal() ? count : ol.getTotal();
		int remaining = totalNeeded - retrieved;
		while (remaining > 0) {
			pageSize = remaining < serverPageSize ? remaining : serverPageSize;
			start = retrieved;
			ObjectList nextList = d1Node.listObjects(null, fromDate, toDate, formatid, replicaStatus, start, pageSize);
			retrieved += nextList.sizeObjectInfoList();
			remaining = totalNeeded - retrieved;
			ol.getObjectInfoList().addAll(nextList.getObjectInfoList());
			logger.info(String.format("start: %d, pageSize: %d, count: %d, total: %d",
					start,
					pageSize,
					nextList.sizeObjectInfoList(),
					nextList.getTotal()
					));
		}
		ol.setCount(ol.sizeObjectInfoList());
		return ol;
    }

	
	/**
	 * Calls list objects iteratively using the paging mechanism in order to
	 * respect any server-imposed paging limits
	 * @param d1Node
	 * @param fromDate
	 * @param toDate
	 * @param formatid
	 * @param replicaStatus
	 * @param start
	 * @param count
	 * @return
	 * @throws InvalidRequest
	 * @throws InvalidToken
	 * @throws NotAuthorized
	 * @throws NotImplemented
	 * @throws ServiceFailure
	 * @throws InsufficientResources 
	 */
	public static Log pagedGetLogRecords(D1Node d1Node, Date fromDate, Date toDate, 
      Event event, String pidFilter, Integer start, Integer count) 
      throws InvalidToken, InvalidRequest, ServiceFailure, NotAuthorized, NotImplemented, InsufficientResources 
    {

		if (count != null && count <= 0) {
			return d1Node.getLogRecords(null, fromDate, toDate, event, pidFilter, start, 0);
		} 
		
		Log entries = d1Node.getLogRecords(null, fromDate, toDate, event, pidFilter, start, null);
		if (entries.getTotal() == entries.sizeLogEntryList()) {
			// don't need to ask for more
			if (count != null && entries.sizeLogEntryList() > count) {
				// need to trim the object list to match the requested amount
				entries.setLogEntryList(entries.getLogEntryList().subList(0, count));
				entries.setCount(count);
			}
		}
		
		count = -1;
		int retrieved = entries.sizeLogEntryList();
		int serverPageSize = entries.sizeLogEntryList();  // server is happy to return this amount at a time.
		int totalNeeded = (count > entries.getTotal()) ? count : entries.getTotal();
		int remaining = totalNeeded - retrieved;
		while (remaining > 0) {
			int pageSize = remaining < serverPageSize ? remaining : serverPageSize;
			start = retrieved;
			Log nextList = d1Node.getLogRecords(null, fromDate, toDate, event, pidFilter, start, pageSize);
			retrieved += nextList.sizeLogEntryList();
			remaining = totalNeeded - retrieved;
			entries.getLogEntryList().addAll(nextList.getLogEntryList());
		}
		entries.setCount(entries.sizeLogEntryList());
		return entries;
    }

	public static ObjectList pagedSolrSearch(AuditCNode cn, D1Url query, Integer start, Integer count) 
	throws InvalidRequest, InvalidToken, NotAuthorized, NotImplemented, ServiceFailure
	{
		String assembledQuery = query.getAssembledQueryString();
		//		D1Url d1url = new D1Url("foo");
//		d1url.addNonEmptyParamPair("q", "*/*");
		query.addNonEmptyParamPair("start",start);
		
		if (count != null && count <= 0) {
			query.addNonEmptyParamPair("rows", 0);
			return cn.search("solr",query);
		} 
		query.addNonEmptyParamPair("rows", 5000);
		ObjectList ol = cn.search("solr",query);
		logger.debug(cn.getLatestRequestUrl());
		
		if (ol.getTotal() == ol.sizeObjectInfoList()) {
			// don't need to ask for more
			if (count != null && ol.sizeObjectInfoList() > count) {
				// need to trim the object list to match the requested amount
				ol.setObjectInfoList(ol.getObjectInfoList().subList(0, count));
				ol.setCount(count);
			}
		}

		count = -1;
		int retrieved = ol.sizeObjectInfoList();
		int serverPageSize = ol.sizeObjectInfoList();  // server is happy to return this amount at a time.
		int totalNeeded = count > ol.getTotal() ? count : ol.getTotal();
		int remaining = totalNeeded - retrieved;
		while (remaining > 0) {
			int pageSize = remaining < serverPageSize ? remaining : serverPageSize;
			start = retrieved;
			D1Url q = new D1Url("foo too");
			q.addPreEncodedNonEmptyQueryParams(assembledQuery);
			q.addNonEmptyParamPair("start", start);
			q.addNonEmptyParamPair("rows", pageSize);
			
			ObjectList nextList = cn.search("solr",q);
			logger.debug(cn.getLatestRequestUrl());
			retrieved += nextList.sizeObjectInfoList();
			remaining = totalNeeded - retrieved;
			ol.getObjectInfoList().addAll(nextList.getObjectInfoList());
		}
		ol.setCount(ol.sizeObjectInfoList());
		return ol;
	}
	
	/**
	 * A recursive method that returns a map of key value pairs that correspond 
	 * to the property paths (keys) and values of the object passed in.
	 * It depends on bean-style getter methods to find the properties of the 
	 * object.
	 * It will sort subtypes that are arrays when order doesn't matter.  For a strict
	 * non-sorting listing, use getD1SubtypesListing(String path, Object o, boolean sortArrays)
	 * and set sortArrays to false.
	 * @param o - the dataone object
	 * @return using LinkedHashMap to ensure order read equals order listed
	 */
	public static LinkedHashMap<String,String> getD1SubtypesListing(Object o) {
		return getD1SubtypesListing(o.getClass().getSimpleName(), o, true);
		
	}
	
	/**
	 * A recursive method that returns a map of key value pairs that correspond 
	 * to the property paths (keys) and values of the object passed in.
	 * It depends on bean-style getter methods to find the properties of the 
	 * object.
	 * Even with sortArrays = true, the Replica list will not be sorted 
	 * because order is considered important. 
	 * @param path
	 * @param o - the dataone object
	 * @param sortArrays - true will sort array subtypes so they can be logically compared
	 * @return using LinkedHashMap to ensure order read equals order listed
	 */
	public static LinkedHashMap<String,String> getD1SubtypesListing(String path, Object o, boolean sortArrays) {
		
		LinkedHashMap<String,String> results = new LinkedHashMap<String,String>();
		if (o == null) {
			results.put(path, "null");
		}
		else {

			Class cl = o.getClass();
			logger.debug("listing for: " + path + ": " + cl.getName());
			// only dataone service types are dealt with here, because we know 
			// they follow the bean convention of set/get methods to access properties
			if (cl.getPackage().getName().startsWith("org.dataone.service.types")) {
				Method[] methods = o.getClass().getMethods();
				for (Method m : methods) {
					// if a get method that isn't "getClass", and doesn't take
					// a parameter (avoids list-associated methods) OR is an
					// enumeration (that uses xmlValue to get the value...
					if ((m.getName().startsWith("get") && !m.getName().equals("getClass")
							&& m.getParameterTypes().length == 0)
							|| m.getName().equals("xmlValue")) {
						try {
							Object p = m.invoke(o, (Object[]) null);
							Map<String,String> subresults = getD1SubtypesListing(
									path + "/" + m.getName().replace("get", ""),
									p,
									sortArrays);
							results.putAll(subresults);
						} catch (IllegalArgumentException e) {
							logger.error(m.getName(),e);
							// TODO Auto-generated catch block
							e.printStackTrace();
						} catch (IllegalAccessException e) {
							logger.error(m.getName(),e);
							// TODO Auto-generated catch block
							e.printStackTrace();
						} catch (InvocationTargetException e) {
							logger.error(m.getName(),e);
							// TODO Auto-generated catch block
							e.printStackTrace();
						}
					} 
				}
			} 
			// otherwise, we are at a more basic datatype (String, etc)
			else {
				if (o instanceof String) {
					results.put(path, (String) o);
				}
				else if (o instanceof Integer) {
					results.put(path, ((Integer)o).toString());
				}
				else if (o instanceof Date) {
					results.put(path, ((Date)o).toString());
				}
				else if (o instanceof BigInteger) {
					results.put(path, ((BigInteger)o).toString());
				}
				else if (o instanceof Boolean) {
					results.put(path, ((Boolean)o).toString());
				} 
				else if (o instanceof ArrayList) {	
					ArrayList<?> list = ((ArrayList<?>) o);
					if ( !list.isEmpty() ) {
						if (sortArrays) {
							logger.debug("arrayList size: " + list.size() );
							// many of the dataone simpletypes are sortable (implement Comparable)
							if (list.get(0) instanceof Comparable) {
								// permissions, Subjects, and Identifiers are Comparable, so sort and go
								Collections.sort((List<Comparable>)list);
							} 
							// ADD COMPARATORS HERE for subtype properties that are
							// lists of non-Comparable dataone service-types
							// if those need to be sorted (where list order doesn't 
							// affect behavior)
							
							// The order of AccessRules in an AccessPolicy doesn't
							// affect behavior, so should be made Comparable so 
							// equivalence can be tested
							else if (list.get(0) instanceof AccessRule) {
								Collections.sort((List<AccessRule>)list, 
								    new Comparator<AccessRule>() {
									    public int compare(AccessRule ar1, AccessRule ar2) {
									    	
									    	// only using subjectlist for sorting
									    	// because it's highly unlikely that 
									    	// that the same subject list will be
									    	// listed twice under separate AccessRules
									    	StringBuffer ar1SB = new StringBuffer();
									    	Collections.sort(ar1.getSubjectList());
									    	for(Subject s : ar1.getSubjectList()) {
									    		ar1SB.append(s.getValue());
									    	}
									    	String ar1Subjects = ar1SB.toString();
									    	
									    	StringBuffer ar2SB = new StringBuffer();
									    	Collections.sort(ar2.getSubjectList());
									    	for(Subject s : ar2.getSubjectList()) {
									    		ar2SB.append(s.getValue());
									    	}
									    	String ar2Subjects = ar2SB.toString();

									    	return ar1Subjects.compareTo(ar2Subjects);
									    }
								    }
								);
							}
							else {
							    ;	// don't sort...
							}
						} 
						for (int j = 0; j < list.size(); j++) {
							results.putAll(getD1SubtypesListing( (path + "/" + j),
						             list.get(j),sortArrays));
						}
					} 
					else {
						results.put(path,"{empty}");
					}
				}
			}
		}
		logger.debug("(return)");
		return results;
	}

	
}
