package org.dataone.client.audit;

import java.io.InputStream;
import java.util.Date;

import org.dataone.client.MNode;
import org.dataone.service.exceptions.InsufficientResources;
import org.dataone.service.exceptions.InvalidRequest;
import org.dataone.service.exceptions.InvalidToken;
import org.dataone.service.exceptions.NotAuthorized;
import org.dataone.service.exceptions.NotFound;
import org.dataone.service.exceptions.NotImplemented;
import org.dataone.service.exceptions.ServiceFailure;
import org.dataone.service.mn.tier1.v1.MNCore;
import org.dataone.service.types.v1.Checksum;
import org.dataone.service.types.v1.DescribeResponse;
import org.dataone.service.types.v1.Event;
import org.dataone.service.types.v1.Identifier;
import org.dataone.service.types.v1.Log;
import org.dataone.service.types.v1.Node;
import org.dataone.service.types.v1.ObjectFormatIdentifier;
import org.dataone.service.types.v1.ObjectList;
import org.dataone.service.types.v1.Permission;
import org.dataone.service.types.v1.SystemMetadata;
import org.dataone.service.util.D1Url;

/**
 * A D1Node subclass that only provides selected methods from the DataONE MN api
 * that will not change the state of the production machines (other than possibly
 * adding log events)
 * 
 * Usage:  AuditMNode mna = new AuditMNode(mnBaseUrl);
 *    or:      D1Node mna = new AuditCNode(baseUrl);
 * 
 * @author rnahf
 *
 */
public class AuditMNode extends D1NodeAuditingWrapper {

	private MNode mn;
	
	public AuditMNode(String mnBaseUrl) {
		super(mnBaseUrl);
		this.mn = new MNode(mnBaseUrl);
	}

	
	public String getNodeBaseServiceUrl() {
		D1Url url = new D1Url(super.getNodeBaseServiceUrl());
		url.addNextPathElement(MNCore.SERVICE_VERSION);
		return url.getUrl();
	}
	
	// set of sessionless api methods that haven't been pushed down into D1Node yet
	// when they do, these "overrides" can be removed from here
	// == these methods are convenience methods that hit the superclass
	
	public ObjectList listObjects() 
	throws InvalidRequest, InvalidToken, NotAuthorized, NotImplemented, ServiceFailure 
	{
		return super.listObjects();
	}
	
	
	public ObjectList listObjects(Date fromDate, Date toDate, ObjectFormatIdentifier formatid, 
			Boolean replicaStatus, Integer start, Integer count) 
	throws InvalidRequest, InvalidToken, NotAuthorized, NotImplemented, ServiceFailure
	{
		return super.listObjects(fromDate, toDate, formatid, replicaStatus, start, count);
	}
	
	
	public SystemMetadata getSystemMetadata(Identifier pid) 
	throws InvalidToken, ServiceFailure, NotAuthorized, NotFound, NotImplemented 
	{
		return super.getSystemMetadata(pid);
	}
	
	/**
	 * uses getReplica() instead to avoid unnecessary log events from auditing
	 * @param pid
	 * @return
	 * @throws InvalidToken
	 * @throws ServiceFailure
	 * @throws NotAuthorized
	 * @throws NotFound
	 * @throws NotImplemented
	 * @throws InsufficientResources
	 */
	public InputStream get(Identifier pid) 
	throws InvalidToken, ServiceFailure, NotAuthorized, NotFound, NotImplemented, InsufficientResources 
	{
		return mn.get(pid);
	}
	
	public InputStream getReplica(Identifier pid) 
	throws InvalidToken, ServiceFailure, NotAuthorized, NotFound, NotImplemented, InsufficientResources 
	{
		return mn.getReplica(pid);
	}
		
	
	public Node getCapabilities() 
	throws NotImplemented, ServiceFailure 
	{
		return mn.getCapabilities();
	}
	
	
	public Checksum getChecksum(Identifier pid, String algorithm) 
	throws InvalidRequest, InvalidToken, NotAuthorized, NotImplemented, ServiceFailure, NotFound 
	{
		return mn.getChecksum(pid, algorithm);
	}
	

	
	public Log getLogRecords(Date fromDate, Date toDate, Event event, 
			String pidFilter, Integer start, Integer count) 
	throws InvalidToken, InvalidRequest, ServiceFailure, NotAuthorized, 
	NotImplemented, InsufficientResources 
	{
		return super.getLogRecords(fromDate, toDate, event, pidFilter, start, count);
	}
	
	public DescribeResponse describe(Identifier pid) 
	throws InvalidToken, NotAuthorized, NotImplemented, ServiceFailure, NotFound
	{
		return super.describe(pid);
	}
	
	
	public boolean isAuthorized(Identifier pid, Permission permission) 
	throws ServiceFailure, InvalidToken, NotFound, NotAuthorized, NotImplemented, InvalidRequest
	{
		return super.isAuthorized(pid, permission);
	}
}
